%{
#include <vector>
#include "states.h"

#define PRE_SIZE(n) {NumSize##n++; CurrentState.Size = n; StateStack.push_back(CurrentState); }
#define POST_SIZE(n)                       \
	{                                  \
	  NumSize##n--; /* Useless var */  \
                                           \
	  std::vector<State>::iterator it; \
                                           \
	  it = StateStack.end();           \
	  do{                              \
	   it--;                           \
	  }while(it->Size == n && it != StateStack.begin());\
	                                   \
	  if(it == StateStack.begin() && it->Size == n)\
	         yyerror("Unmatched </" #n "> found!\n");\
                                           \
          CurrentState.Size = it->Size;\
	  StateStack.push_back(CurrentState);\
	}

#define PRE_COLOR(c) { Num##c++; ColorStack.push_back(CurrentState.Color); CurrentState.Color = c; StateStack.push_back(CurrentState); }
#define POST_COLOR(c) \
        {\
	    CurrentState.Color = ColorStack.back();\
	    ColorStack.pop_back();\
	    StateStack.push_back(CurrentState);\
	}

int yylex();
void yyerror(char *message);
extern char *yytext;

extern std::vector<State> StateStack;
static State CurrentState;

static std::vector<Colors> ColorStack;

static int NumBold=0;
static int NumEmphasis=0;
static int NumStrongEmphasis=0;
static int NumItalics=0;
static int NumTypewriter=0;
static int NumPlain=0;
static int NumUnderline=0;

static int NumRed=0;
static int NumGreen=0;
static int NumBlue=0;
static int NumCyan=0;
static int NumMagenta=0;
static int NumYellow=0;
static int NumBlack=0;
static int NumWhite=0;

static int NumSize0=0;
static int NumSize1=0;
static int NumSize2=0;
static int NumSize3=0;
static int NumSize4=0;
static int NumSize5=0;
static int NumSize6=0;
static int NumSize7=0;
static int NumSize8=0;
static int NumSize9=0;

%}

%token BOLD
%token C_BOLD
%token EMPHASIS
%token C_EMPHASIS
%token ITALICS
%token C_ITALICS
%token PLAIN
%token C_PLAIN
%token STRONG_EMPHASIS
%token C_STRONG_EMPHASIS
%token TYPEWRITER
%token C_TYPEWRITER
%token UNDERLINE
%token C_UNDERLINE
%token SIZE_0
%token C_SIZE_0
%token SIZE_1
%token C_SIZE_1
%token SIZE_2
%token C_SIZE_2
%token SIZE_3
%token C_SIZE_3
%token SIZE_4
%token C_SIZE_4
%token SIZE_5
%token C_SIZE_5
%token SIZE_6
%token C_SIZE_6
%token SIZE_7
%token C_SIZE_7
%token SIZE_8
%token C_SIZE_8
%token SIZE_9
%token C_SIZE_9
%token RED
%token C_RED
%token GREEN
%token C_GREEN
%token BLUE
%token C_BLUE
%token CYAN
%token C_CYAN
%token MAGENTA
%token C_MAGENTA
%token YELLOW
%token C_YELLOW
%token BLACK
%token C_BLACK
%token WHITE
%token C_WHITE
%token TEXT

%%

file: document { CurrentState.Text = ""; StateStack.push_back(CurrentState); }
    | /* blank to trap empty files */

document: document document
        | TEXT { CurrentState.Text = yytext; StateStack.push_back(CurrentState); CurrentState.Text = ""; }

        | BOLD { NumBold++; CurrentState.Bold = true; StateStack.push_back(CurrentState); }
          document C_BOLD { if(!--NumBold) CurrentState.Bold = false; StateStack.push_back(CurrentState); }

        | EMPHASIS { NumEmphasis++; if(!CurrentState.StrongEmphasis) CurrentState.Emphasis = !CurrentState.Emphasis; StateStack.push_back(CurrentState); }
          document C_EMPHASIS { NumEmphasis--; if(!CurrentState.StrongEmphasis) CurrentState.Emphasis = !CurrentState.Emphasis; StateStack.push_back(CurrentState); }

        | ITALICS { NumItalics++; CurrentState.Italics = true; StateStack.push_back(CurrentState); }
          document C_ITALICS { if(!--NumItalics) CurrentState.Italics = false; StateStack.push_back(CurrentState); }

        | PLAIN { NumPlain++;
                  CurrentState.Plain = true;
                  CurrentState.Underline = 0;
                  CurrentState.Bold =
                  CurrentState.Emphasis =
                  CurrentState.StrongEmphasis =
                  CurrentState.Italics =
                  CurrentState.Typewriter = false;
		  StateStack.push_back(CurrentState); 
                }
          document C_PLAIN
                {
                   std::vector<State>::iterator it;

                   it = StateStack.end();
                   do
                   {
                     it--;
                   }while(it->Plain && it != StateStack.begin());

                   CurrentState.Plain = false;
                   CurrentState.Underline = it->Underline;
                   CurrentState.Bold = it->Bold;
                   CurrentState.Emphasis = it->Emphasis;
                   CurrentState.Italics = it->Italics;
                   CurrentState.Typewriter = it->Typewriter;
		   StateStack.push_back(CurrentState); 
                }

        | STRONG_EMPHASIS { NumStrongEmphasis++; CurrentState.StrongEmphasis = true; StateStack.push_back(CurrentState); }
          document C_STRONG_EMPHASIS { if(!--NumStrongEmphasis) CurrentState.StrongEmphasis = false; StateStack.push_back(CurrentState); }

        | TYPEWRITER { NumTypewriter++; CurrentState.Typewriter = true; StateStack.push_back(CurrentState); }
          document C_TYPEWRITER { if(!--NumTypewriter) CurrentState.Typewriter = false; StateStack.push_back(CurrentState); }

        | UNDERLINE { NumUnderline++; if(CurrentState.Underline < 3) CurrentState.Underline++; StateStack.push_back(CurrentState); }
          document C_UNDERLINE { if(--NumUnderline <= 2) CurrentState.Underline--; StateStack.push_back(CurrentState); }

        | SIZE_0 { PRE_SIZE(0); } 
          document C_SIZE_0 { POST_SIZE(0); }

        | SIZE_1 { PRE_SIZE(1); }
          document C_SIZE_1 { POST_SIZE(1); }
	 
	| SIZE_2 { PRE_SIZE(2); }
          document C_SIZE_2 { POST_SIZE(2); } 
	   
	| SIZE_3 { PRE_SIZE(3); }
	  document C_SIZE_3 { POST_SIZE(3); }
	  
	| SIZE_4 { PRE_SIZE(4); }
	  document C_SIZE_4 { POST_SIZE(4); }
	  
	| SIZE_5 { PRE_SIZE(5); }
	  document C_SIZE_5 { POST_SIZE(5); }
	  
	| SIZE_6 { PRE_SIZE(6); }
	  document C_SIZE_6 { POST_SIZE(6); }
	  
	| SIZE_7 { PRE_SIZE(7); }
	  document C_SIZE_7 { POST_SIZE(7); }

	| SIZE_8 { PRE_SIZE(8); }
	  document C_SIZE_8 { POST_SIZE(8); }
	  
	| SIZE_9 { PRE_SIZE(9); }
	  document C_SIZE_9 { POST_SIZE(9); }

	| RED { PRE_COLOR(Red); }
	  document C_RED { POST_COLOR(Red); }

	| GREEN { PRE_COLOR(Green); }
	  document C_GREEN { POST_COLOR(Green); }

	| BLUE { PRE_COLOR(Blue); }
	  document C_BLUE { POST_COLOR(Blue); }

	| CYAN { PRE_COLOR(Cyan); }
	  document C_CYAN { POST_COLOR(Cyan); }

	| MAGENTA { PRE_COLOR(Magenta); }
	  document C_MAGENTA { POST_COLOR(Magenta); }
 
	| YELLOW { PRE_COLOR(Yellow); }
	  document C_YELLOW { POST_COLOR(Yellow); }
 
	| BLACK { PRE_COLOR(Black); }
	  document C_BLACK { POST_COLOR(Black); }

	| WHITE { PRE_COLOR(White); }
	document C_WHITE { POST_COLOR(White); }

	/* trap junk data */
	| BOLD C_BOLD
	| EMPHASIS C_EMPHASIS
	| ITALICS C_ITALICS
	| PLAIN C_PLAIN
	| STRONG_EMPHASIS C_STRONG_EMPHASIS
	| TYPEWRITER C_TYPEWRITER
	| UNDERLINE C_UNDERLINE
	| SIZE_0 C_SIZE_0
	| SIZE_1 C_SIZE_1
	| SIZE_2 C_SIZE_2
	| SIZE_3 C_SIZE_3
	| SIZE_4 C_SIZE_4
	| SIZE_5 C_SIZE_5
	| SIZE_6 C_SIZE_6
	| SIZE_7 C_SIZE_7
	| SIZE_8 C_SIZE_8
	| SIZE_9 C_SIZE_9
	| RED C_RED
	| GREEN C_GREEN
	| BLUE C_BLUE
	| CYAN C_CYAN
	| MAGENTA C_MAGENTA
	| YELLOW C_YELLOW
	| BLACK C_BLACK
	| WHITE C_WHITE

%%

void yyerror(char *Message)
{
  printf("Parser Error: %s\n", Message);
  exit(-1);
}

